/*********************************
 * This file is part of the "Interactive Fan Chart Demo"
 * by Geoffrey M. Draper.
 * 
 * Geoffrey M. Draper, author of this work, hereby grants
 * everyone permission to use and distribute this code,
 * with or without modifications for any purpose,
 * commercial or otherwise. 
 * 
 * It is requested, although not required, that derivative
 * works include an acknowledgment somewhere in the
 * derivative work's credits.
 *********************************/

package fanchart;




import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Point2D;
import java.io.Serializable;
import java.util.Set;

import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.Timer;
import javax.swing.event.MouseInputAdapter;

import fanchart.Ward.WardListener;


public class FanChartWidget extends JPanel implements WardListener, ActionListener {
	private Ward ward;
	private PersonWedge rootWedge;
	protected Set<Person> excludeFromRing;
	private static final long serialVersionUID = 42L;
	protected RadialCache cache;
	protected Dimension windowSize;
	protected boolean dirty;
	protected float innerRadius, outerRadius;
	protected float[] radii;
	protected Point center;
	protected int spinFactor;
	protected boolean spinning;
	private boolean highlightRootNode;
	private GotoChildButton centerButton;
	private PersonWedge wedgeClickedOn = null;
	private float animationRatio;
	public static boolean currentlyAnimating;
	public static boolean currentlyAnimatingNewRoot;
	private Timer timer;//moving this to class level to satisfy compiler
	protected DartboardMouseHandler mouse;
	private HyrumPolygon draggedPolygon;
	private final static String EDIT_PERSON_COMMAND = "EDIT_PERSON_COMMAND";
	private final static String HIDE_WEDGE_COMMAND = "HIDE_WEDGE_COMMAND";
	private final static String NEW_ROOT_CHILD_COMMAND_PREFIX = "CHILD_";
	private final static String NEW_ROOT_COMMAND_PREFIX = "NEWROOT_";
	
	private enum AnimationStyle {
		GROW,
		SHRINK
	}

	protected FanChartWidget() {
		super();
		cache = new RadialCache();
		spinFactor = RadialCache.RESOLUTION/2;
		spinning = false;
		dirty = false;
		currentlyAnimating = false;
		currentlyAnimatingNewRoot = false;
		highlightRootNode = false;
		draggedPolygon = null;
	}

	public FanChartWidget(Ward ward) {
		this();
		this.ward = ward;
		this.ward.addWardListener(this);
		mouse = new DartboardMouseHandler();
		addMouseListener(mouse);
		addMouseMotionListener(mouse);
		init(null, null);
	}

	private void init(PersonWedge fw, PersonWedge mw) {
		rootWedge = new PersonWedge(ward.getRootPerson(), 0, RadialCache.RESOLUTION, 0, fw, mw);
		rootWedge.setAsRoot(true);
		centerButton = new GotoChildButton();
		if (rootWedge.getEntity().hasChild()) {
			centerButton.setVisible(true);
		}
		resize();
		resizeWedges();		
	}

	protected void resize() {
		windowSize = getSize();
		innerRadius = Config.rootNodeRadius;
		
		//guard against ArrayIndexOutOfBoundsExceptions
		if (PersonWedge.getHighestVisibleGeneration() >= Config.db_totalNumberOfRadii) {
			Config.db_totalNumberOfRadii += 2;
		}

		radii = new float[Config.db_totalNumberOfRadii];
		radii[0] = innerRadius;
		radii[1] = innerRadius+30;//magic number :(
		for (int i=2; i<Config.db_totalNumberOfRadii; ++i) {
			//Use this line if you want ring-spacing similar
			//to the one in the FHTW2008 demo.
			radii[i] = radii[i-1] + Config.getWedgeThicknessForGeneration(i);
			//OR
			//Use this one if you want ring-spacing similar
			//to PAF Companion's.
			//radii[i] = Config.getNextWedgeThickness(radii[i-1]);
		}

		center = new Point(windowSize.width/2, windowSize.height/2);
		cache.setRadii(radii, center);
		if (centerButton != null) centerButton.refresh();
		dirty = true;
	}

	protected void resizeWedges() {
		rootWedge.buildVertices(cache, windowSize, spinFactor);
	}

	public void wardChanged() {
		init(null, null);
		repaint();
	}

	@Override
	public void paintComponent(Graphics g1) {
		Graphics2D g = (Graphics2D)g1;
		if (!(windowSize.equals(getSize()))) {
			resize();
		}
		if (dirty) {
			resizeWedges();
			dirty = false;
		}

		//turn on antialiasing, if desired.
		if (Config.antialias && !(spinning || currentlyAnimating)) {
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		} else {
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
		}

		CurvedText textRenderer = new CurvedText(cache, windowSize);

		//set the line thickness
		g.setStroke(Config.regularLine);

		//Clear screen to background color
		g.setColor(Config.backgroundColor);
		g.fillRect(0,0,windowSize.width, windowSize.height);

		rootWedge.draw(textRenderer, g, !spinning);
		rootWedge.drawThickOutlineForExpandedWedges(g);
		if (highlightRootNode) {
			rootWedge.drawThickOutline(g);
		}

		if (!currentlyAnimatingNewRoot) {			
			centerButton.draw(g);
		}
		
		if (draggedPolygon != null)  {
			g.setColor(Color.ORANGE);
			g.setStroke(Config.thinLine);
			draggedPolygon.draw(g);
		}


	}

	protected class DartboardMouseHandler extends MouseInputAdapter implements Serializable {
		private boolean dragging = false;
		private Point prevMousePos, mousePos;
		private Point draggedPolygonOriginalPos;//TODO for future use
		private static final long serialVersionUID = 42L;

		@Override
		public void mousePressed(MouseEvent e) {
			//process no mouse events if we're animating.
			if (currentlyAnimating) return;

			mousePos = e.getPoint();
			prevMousePos = mousePos;
			//figure out which wedge was clicked on
			wedgeClickedOn = whichWedgeIsMouseIn();
			if (e.getButton() == 1) {
				if (wedgeClickedOn != null) {
					if (wedgeClickedOn.isRoot()) {
						if (centerButton.containsPoint(mousePos)) {
							setOneOfTheChildrenAsTheNewRoot(mousePos);
						}
					} else {
						dragging = true;
					}
				}
			} else {
				if (wedgeClickedOn == null) {
					spinning = true;
				} else {
					showPopupMenu(mousePos);
				}
			}
		}

		private PersonWedge whichWedgeIsMouseIn() {
			return rootWedge.findClickedWedge(mousePos);
		}

		@Override
		public void mouseReleased(MouseEvent e) {
			//process no mouse events if we're animating.
			if (currentlyAnimating) return;

			mousePos = e.getPoint();
			if (dragging) {
				//figure out which wedge was clicked on
				PersonWedge pw = whichWedgeIsMouseIn();
				if (pw != null) {
					//mouse was released inside chart, so...
					if (pw != rootWedge) {
						//if we clicked anywhere but the root wedge...
						//if this was a simple click (without dragging)
						//then expand or contract the current wedge.
						if (pw == wedgeClickedOn) {
							if (draggedPolygon == null) {
								if (pw.isPointInsideForceGenerationWedge(mousePos)) {
									pw.forceNextGeneration();
									resize();
								} else {
									expandOrContract(pw);
								}
							}
						}
						//for anything other than simple clicks
						//outside of the root note, fly home.
						animateDraggedPolygonBackHome();
					} else {
						//if we released the mouse on the root node,
						//and if we dragged here from somewhere else,
						//then make the dragged node the root node.
						if (pw != wedgeClickedOn) {
							setNewRoot(wedgeClickedOn.getPerson());
						}
						draggedPolygon = null;
					}
				} else {
					//mouse was released outside of chart, fly home.
					animateDraggedPolygonBackHome();
				}
			}
			spinning = false;
			dragging = false;
			highlightRootNode = false;
			repaint();
		}

		//TODO add animation
		private void animateDraggedPolygonBackHome() {
			draggedPolygon = null;
		}

		@Override
		public void mouseDragged(MouseEvent e) {
			//process no mouse events if we're animating.
			if (currentlyAnimating) return;

			mousePos = e.getPoint();			
			Point delta = new Point(mousePos.x-prevMousePos.x,
					mousePos.y-prevMousePos.y);
			if (spinning) {
				//translate the user's mouse movement into
				//a direction of rotation for the circle.
				int absX = Math.abs(delta.x);
				int absY = Math.abs(delta.y);
				if (mousePos.x > center.x) {
					if (mousePos.y > center.y) {
						if (absX > absY) spinFactor -= delta.x;
						else spinFactor += delta.y;
					} else {
						if (absX > absY) spinFactor += delta.x;
						else spinFactor += delta.y;
					}
				} else {
					if (mousePos.y > center.y) {
						if (absX > absY) spinFactor -= delta.x;
						else spinFactor -= delta.y;
					} else {
						if (absX > absY) spinFactor += delta.x;
						else spinFactor -= delta.y;
					}					
				}
				spinFactor %= RadialCache.RESOLUTION;
				dirty = true;
			}
			if (dragging) {
				if (draggedPolygon == null) {
					draggedPolygon = wedgeClickedOn.copyPolygon();
					draggedPolygonOriginalPos = prevMousePos;
				}
				draggedPolygon.transform(AffineTransform.getTranslateInstance(delta.x, delta.y));
				PersonWedge pw = whichWedgeIsMouseIn();
				highlightRootNode = false;
				if (pw != null) {
					if (pw == rootWedge) {
						if (pw != wedgeClickedOn) {
							highlightRootNode = true;
						}
					}
				}

			}
			prevMousePos = mousePos;
			repaint();
		}

		@Override
		public void mouseMoved(MouseEvent e) {
			//process no mouse events if we're animating.
			if (currentlyAnimating) {
				FanChartWidget.this.setToolTipText(null);
				return;
			}

			try {
				mousePos = e.getPoint();
				String tooltip = null;
				if (centerButton.containsPoint(mousePos)) {
					tooltip = "Click here to go back one generation.";
					centerButton.setHighlighted(true);
					repaint();
				} else {
					//we only issue a repaint if we're switching
					//from highlighted to non-highlighted.
					if (centerButton.setHighlighted(false)) {
						repaint();
					}
					PersonWedge fw = whichWedgeIsMouseIn();
					if (fw != null) {
						tooltip = fw.getTooltipText();
					}
				}

				FanChartWidget.this.setToolTipText(tooltip);
			} catch (Exception npe) {
				//do nothing; these mouse move events come fast
				//enough that if one chokes, we can skip it.
			}
		}
	}
	
	private void setOneOfTheChildrenAsTheNewRoot(Point pos) {
		if (rootWedge.getEntity().hasMultipleChildren()) {
			showChooseChildMenu(pos);
		} else {
			setNewRoot(rootWedge.getEntity().getChild(), rootWedge);
		}
	}

	private void expandOrContract(PersonWedge pw) {
		PersonWedge spouse = pw.getSpouseWedge();
		if (pw.isExpanded()) {
			if (Config.db_animate) {
				initiateAnimation(pw, AnimationStyle.SHRINK);
			} else {
				pw.setNotExpanded();
				pw.reset();
				spouse.setCollapsed(false);
				spouse.reset();

			}
		} else {
			PersonWedge wedgeToManipulate = spouse;
			PersonWedge otherWedge = pw;
			if (Config.clickToExpand) {
				wedgeToManipulate = pw;
				otherWedge = spouse;
			}
			if (Config.db_animate) {
				initiateAnimation(wedgeToManipulate, AnimationStyle.GROW);
			} else {
				wedgeToManipulate.setExpanded();
				otherWedge.setCollapsed(true);
			}
		}
		dirty = true;
		repaint();
	}

	public void showChooseChildMenu(Point pos) {
		JPopupMenu menu = new JPopupMenu();
		JMenuItem popupOption;
		for (int i=0; i<rootWedge.getEntity().getNumChildren(); ++i) {
			Person child = rootWedge.getEntity().getChild(i);
			popupOption = new JMenuItem(child.getName());
			popupOption.setActionCommand(NEW_ROOT_CHILD_COMMAND_PREFIX + i);
			popupOption.addActionListener(this);
			menu.add(popupOption);
		}
		menu.show(this, pos.x, pos.y);
	}

	//TODO this method is overly complex.  Is there a way we can make it work
	//without defining two separate Timer classes?
	protected void initiateAnimation(final PersonWedge animatedWedge, final AnimationStyle style) {
		//Kick out the trivial case.
		if (animatedWedge == null) return;

		int delay = Config.db_millisecondsBetweenWedgeAnimationFrames;
		currentlyAnimating = true;
		final PersonWedge spouse = animatedWedge.getSpouseWedge();
		if (style == AnimationStyle.SHRINK) {
			//TODO maybe make increment adjustable, based on the speed of the CPU?
			animationRatio = 2f;
			final float increment = -0.1f;
			spouse.setCollapsed(false);
			timer = new Timer(delay, new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					animationRatio += increment;
					animateWedges(animatedWedge, spouse);
					if (animationRatio < 1.111f) {
						timer.stop();
						animatedWedge.setNotExpanded();
						animatedWedge.reset();
						spouse.setCollapsed(false);
						spouse.reset();
						currentlyAnimating = false;
					}
				}
			});
		} else {
			animationRatio = 1f;
			final float increment = 0.1f;
			timer = new Timer(delay, new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					animationRatio += increment;
					animateWedges(animatedWedge, spouse);
					if (animationRatio > 1.89f) {
						timer.stop();
						spouse.setCollapsed(true);
						animatedWedge.setExpanded();
						currentlyAnimating = false;
					}
				}
			});
		}
		timer.setInitialDelay(0);
		timer.start();			
	}

	private void animateWedges(PersonWedge animatedWedge, PersonWedge spouseWedge) {
		float newLengthForAnimatedWedge;
		float delta;

		//reset everything to its 'normal' size.
		animatedWedge.reset();
		spouseWedge.reset();

		newLengthForAnimatedWedge = animatedWedge.getLength() * animationRatio;
		delta = newLengthForAnimatedWedge - animatedWedge.getLength();
		animatedWedge.resetLength(newLengthForAnimatedWedge);
		spouseWedge.adjustLengthByDelta(-delta);
		float start1 = animatedWedge.getStart();
		float start2 = spouseWedge.getStart();
		if (start1 > start2) {
			animatedWedge.resetStart(start1-delta);
		} else {
			spouseWedge.resetStart(start2+delta);
		}

		animatedWedge.createAncestralWedges();
		spouseWedge.createAncestralWedges();

		dirty=true;
		repaint();
	}

	private void setNewRoot(Person person) {
		ward.setRootPerson(person);
		init(null, null);
	}
	
	public void setNewRoot(Person rootPerson, PersonWedge previousRoot) {
		if (Config.db_animate) {
			animateNewRoot(rootPerson, previousRoot);
		} else {
			ward.setRootPerson(rootPerson);
			Person parent = previousRoot.getPerson();
			if (rootPerson.getMother().equals(parent)) {
				init(null, previousRoot);
			}
			if (rootPerson.getFather().equals(parent)) {
				init(previousRoot, null);
			}
		}
	}



	private void animateNewRoot(final Person rootPerson, final PersonWedge previousRoot) {
		ward.setRootPerson(rootPerson);
		Person parent = previousRoot.getPerson();
		if (rootPerson.getMother().equals(parent)) {
			init(null, previousRoot);
		}
		if (rootPerson.getFather().equals(parent)) {
			init(previousRoot, null);
		}
		previousRoot.setExpanded();
		previousRoot.getSpouseWedge().setCollapsed(true);

		int delay = Config.db_millisecondsBetweenWedgeAnimationFrames;
		currentlyAnimating = true;
		currentlyAnimatingNewRoot = true;
		final int steps = 10;
		float startingRadius1 = GotoChildButton.RADIUS;
		final float endingRadius1 = radii[1];
		
		//compute increments for each radius
		final float increments[] = new float[Config.db_totalNumberOfRadii-1];
		float tmp[] = new float[Config.db_totalNumberOfRadii];
		increments[0] = (endingRadius1-startingRadius1)/steps;
		for (int i=1; i<increments.length; ++i) {
			increments[i] = (radii[i+1]-radii[i])/steps;
		}
		
		for (int i=1; i<Config.db_totalNumberOfRadii; ++i) {
			tmp[i] = radii[i-1];
		}
		for (int i=1; i<Config.db_totalNumberOfRadii; ++i) {
			radii[i] = tmp[i];
		}
		radii[0] = 0;
		radii[1] = startingRadius1;
		
		
		timer = new Timer(delay, new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				for (int i=1; i<Config.db_totalNumberOfRadii; ++i) {
					radii[i] += increments[i-1];
				}
				cache.setRadii(radii, center);
				dirty = true;
				repaint();
				if (radii[1] > 0.9f*endingRadius1) {
					timer.stop();
					currentlyAnimatingNewRoot = false;
					currentlyAnimating = false;
					resize();
					expandOrContract(previousRoot);
				}
			}
		});
		timer.setInitialDelay(0);
		timer.start();			
	
	}

	private Graphics2D getGraphics2D() {
		return (Graphics2D)getGraphics();
	}

	private class GotoChildButton {
		private static final float RADIUS = 10;
		private Ellipse2D.Float circle;
		private boolean visible;
		private boolean highlighted;

		GotoChildButton() {
			circle = new Ellipse2D.Float();
			visible = false;
			highlighted = false;
		}

		public void setVisible(boolean b) {
			visible = b;
		}

		public boolean setHighlighted(boolean b) {
			boolean priorValue = highlighted;
			highlighted = b;
			return priorValue;
		}

		public void refresh() {
			refresh(RADIUS);
		}

		public void refresh(float radius) {
			circle.setFrame(center.x-radius, center.y-radius, radius*2, radius*2);
		}

		public boolean containsPoint(Point2D pos) {
			if (visible) {
				return circle.contains(pos);
			} else {
				return false;
			}
		}

		public void draw(Graphics2D g) {
			if (visible) {
				g.setColor(Config.getColorForPerson(ward.getRootPerson().getChild()));
				g.fill(circle);
				g.setColor(Config.wedgeBorderColor);
				if (highlighted) {
					g.setStroke(Config.thickLine);
				} else {
					g.setStroke(Config.regularLine);
				}
				g.draw(circle);
			}
		}
	}

	private void showPopupMenu(Point pos) {
		JPopupMenu menu = new JPopupMenu();
		JMenuItem popupOption = new JMenuItem("Edit...");
		popupOption.setActionCommand(EDIT_PERSON_COMMAND);
		popupOption.addActionListener(this);
		menu.add(popupOption);
		if (!(wedgeClickedOn.isRoot())) {
			if (!(wedgeClickedOn.isCollapsed())) {
				if (Config.clickToExpand) {
					popupOption = new JMenuItem("Expand");
				} else {
					popupOption = new JMenuItem("Hide");
				}
			}
			if (wedgeClickedOn.isExpanded()) {
				popupOption = new JMenuItem("Show spouse");
			}
			popupOption.setActionCommand(HIDE_WEDGE_COMMAND);
			popupOption.addActionListener(this);
			menu.add(popupOption);			
		}
		menu.show(this, pos.x, pos.y);
	}
	
	public void actionPerformed(ActionEvent e) {
		String cmd = e.getActionCommand();
		Person currentPerson = wedgeClickedOn.getPerson();
		Person clone = new Person(currentPerson);
		if (cmd.equals(EDIT_PERSON_COMMAND)) {
			new EditPersonDialog(Main.instance(), true, clone);
			if (clone.isChanged()) {
				currentPerson.setName(clone.getName());
				currentPerson.setBirthdate(clone.getBirthdate());
				repaint();
			}
		}
		if (cmd.equals(HIDE_WEDGE_COMMAND)) {
			expandOrContract(wedgeClickedOn);
		}
		if (cmd.startsWith(NEW_ROOT_CHILD_COMMAND_PREFIX)) {
			String substring = cmd.substring(NEW_ROOT_CHILD_COMMAND_PREFIX.length(), cmd.length());
			int childIndex = Integer.parseInt(substring);
			setNewRoot(rootWedge.getEntity().getChild(childIndex), rootWedge);
		}
		if (cmd.startsWith(NEW_ROOT_COMMAND_PREFIX)) {
			String substring = cmd.substring(NEW_ROOT_COMMAND_PREFIX.length(), cmd.length());
			int index = Integer.parseInt(substring);
			setNewRoot(ward.getArbitaryPerson(index));
			repaint();
		}
	}


}
