/*********************************
 * This file is part of the "Interactive Fan Chart Demo"
 * by Geoffrey M. Draper.
 * 
 * Geoffrey M. Draper, author of this work, hereby grants
 * everyone permission to use and distribute this code,
 * with or without modifications for any purpose,
 * commercial or otherwise. 
 * 
 * It is requested, although not required, that derivative
 * works include an acknowledgment somewhere in the
 * derivative work's credits.
 *********************************/

package fanchart;


import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

//This composite class gives me the precision of the new Path2D
//class, while still providing the low-level access to individual
//vertices I got accustomed to when using the legacy Polygon class.

//NOTE: I backported from Path2D.Float to GeneralPath to maintain
//compatibility with Java 1.5.

public class HyrumPolygon implements Serializable {
	private GeneralPath polygon;
	private List<Point2D.Float> points;
	private static final int d = 4;
	
	public HyrumPolygon() {
		polygon = new GeneralPath(GeneralPath.WIND_EVEN_ODD);
		points = new ArrayList<Point2D.Float>();
	}
	
	public HyrumPolygon(HyrumPolygon hp) {
		polygon = new GeneralPath(hp.polygon);
		points = new ArrayList<Point2D.Float>(hp.points);
	}
	
	public void addPoint(final Point2D.Float p) {
		addPoint(p.x, p.y);
	}
	
	public void addPoint(final float x, final float y) {
		if (points.isEmpty()) {
			polygon.moveTo(x, y);
		} else {
			polygon.lineTo(x, y);
		}
		points.add(new Point2D.Float(x,y));
	}
	
	public Point2D.Float getPointAt(final int index) {
		return points.get(index);
	}
	
	public int getNumPoints() {
		return points.size();
	}
	
	//trivial delegate methods
	public void closePath() {
		if (points.size() > 2) {
			polygon.closePath();
		}
	}
	
	public void reset() {
		polygon.reset();
		points.clear();
	}
	
	public void fill(final Graphics2D g) {
		g.fill(polygon);
	}
	
	public void draw(final Graphics2D g) {
		g.draw(polygon);
	}
	
	public void transform(final AffineTransform at) {
		polygon.transform(at);
	}
	
	public boolean contains(final Point2D p) {
		return polygon.contains(p);
	}
	
	public boolean isInsideRadiusSquared(final Point2D center, final int threshold) {
		for (Point2D.Float p : points) {
			if (center.distanceSq(p) < threshold) {
				return true;
			}
		}
		return false;
	}
	
	public void fisheyeWarp(final HyrumPolygon ref, final Point2D center, final int threshold) {
		this.reset();
		for (Point2D.Float p : ref.points) {
			double dist = center.distance(p);
			if (dist < threshold) {
				//formula based on Sarkar and Brown's CHI 1992 paper:
				double newDist = threshold * ((d+1) * (dist/threshold)) / (d*dist/threshold + 1);
				double ratio = newDist / dist;
				this.addPoint(new Point2D.Float((float)(center.getX() + (ratio * (p.x - center.getX()))),
									 (float)(center.getY() + (ratio * (p.y - center.getY())))));
			} else {
				this.addPoint(p);
			}
		}
		this.closePath();
	}


}
